# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""NDS2 data query routines for the TimeSeries
"""

import operator
import sys
import warnings
from math import ceil

from six.moves import (reduce, StringIO)

from astropy.utils.console import ProgressBarOrSpinner

from ...io import nds2 as io_nds2
from ...segments import (Segment, SegmentList)
from ...utils import gprint
from .. import (TimeSeries)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def print_verbose(*args, **kwargs):
    """Utility to print something only if verbose=True is given
    """
    if kwargs.pop('verbose', False):
        gprint(*args, **kwargs)


def _parse_nds_enum_dict_param(channels, key, value):
    if key == 'type':
        enum = io_nds2.Nds2ChannelType
        default = enum.any() - enum.ONLINE.value
    else:
        enum = io_nds2.Nds2DataType
        default = enum.any()
    # set default
    if value is None:
        value = default
    # parse non-int enum representation
    if not isinstance(value, (dict, int)):
        value = enum.find(value).value
    # return dict of ints
    if isinstance(value, int):
        return dict((c, value) for c in channels)
    # here we know ``value`` is a dict, so just fill in the blanks
    value = value.copy()
    for chan in channels:
        value.setdefault(chan, default)
    return value


def set_parameter(connection, parameter, value, verbose=False):
    """Set a parameter for the connection, handling errors as warnings
    """
    value = str(value)
    try:
        if not connection.set_parameter(parameter, value):
            raise ValueError("invalid parameter or value")
    except (AttributeError, ValueError) as exc:
        warnings.warn(
            'failed to set {}={!r}: {}'.format(parameter, value, str(exc)),
            io_nds2.NDSWarning)
    else:
        if verbose:
            print('    [{}] set {}={!r}'.format(connection.get_host(),
                                                parameter, value))


@io_nds2.open_connection
def fetch(channels, start, end, type=None, dtype=None, allow_tape=None,
          connection=None, host=None, port=None, pad=None, verbose=False,
          series_class=TimeSeries):
    # host and port keywords are used by the decorator only
    # pylint: disable=unused-argument
    """Fetch a dict of data series from NDS2

    This method sits underneath `TimeSeries.fetch` and related methods,
    and isn't really designed to be called directly.
    """
    # set ALLOW_DATA_ON_TAPE
    if allow_tape is not None:
        set_parameter(connection, 'ALLOW_DATA_ON_TAPE', str(allow_tape),
                      verbose=verbose)

    type = _parse_nds_enum_dict_param(channels, 'type', type)
    dtype = _parse_nds_enum_dict_param(channels, 'dtype', dtype)

    # verify channels exist
    print_verbose("Checking channels list against NDS2 database...", end=' ',
                  verbose=verbose)
    utype = reduce(operator.or_, type.values())  # logical OR of types
    udtype = reduce(operator.or_, dtype.values())
    ndschannels = io_nds2.find_channels(channels, connection=connection,
                                        type=utype, dtype=udtype, unique=True,
                                        epoch=(start, end))
    names = ['%s,%s' % (c.name, c.channel_type_to_string(c.channel_type)) for
             c in ndschannels]
    print_verbose('done', verbose=verbose)

    # handle minute trend timing
    if (any(c.endswith('m-trend') for c in names) and
            (start % 60 or end % 60)):
        warnings.warn("Requested at least one minute trend, but "
                      "start and stop GPS times are not multiples of "
                      "60. Times will be expanded outwards to compensate")
        start, end = io_nds2.minute_trend_times(start, end)

    # get data availability
    span = SegmentList([Segment(start, end)])
    if pad is None:
        qsegs = span
        gap = None
    else:
        print_verbose("Querying for data availability...", end=' ',
                      verbose=verbose)
        gap = 'pad'
        qsegs = io_nds2.get_availability(
            ndschannels, start, end, connection=connection).intersection()
        qsegs &= span
        print_verbose('done\nFound {0} viable segments of data with {1}%% '
                      'coverage'.format(len(qsegs),
                                        abs(qsegs) / abs(span) * 100),
                      verbose=verbose)
        if span - qsegs:
            warnings.warn("Gaps were found in data available from {0}, "
                          "but will be padded with {1}".format(
                              connection.get_host(), pad))

    # query for each segment
    out = series_class.DictClass()
    for seg in qsegs:
        duration = seg[1] - seg[0]
        msg = 'Downloading data ({}-{} | {}s):'.format(
            seg[1], seg[0], duration)
        stream = sys.stdout if verbose else StringIO()
        count = 0
        with ProgressBarOrSpinner(duration, msg, file=stream) as bar:
            for buffers in connection.iterate(int(seg[0]), int(seg[1]), names):
                for buffer_, chan in zip(buffers, channels):
                    series = series_class.from_nds2_buffer(buffer_)
                    out.append({chan: series}, pad=pad, gap=gap)
                count += buffer_.length / buffer_.channel.sample_rate
                bar.update(count)

    return out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Monkeypatch for numpydoc to plot gwpy docstring examples

By default (as of v0.6), numpydoc will only auto-plot code blocks in
_Examples_ sections if they include 'import matplotlib'. Since GWpy examples
typically don't need to import matplotlib, we replace the relevant method
here with one that detects gwpy's typical plot generation and display code,
namely just `plot.show()`.

Hopefully numpydoc will support other key phrases as of v0.7, so this
module could then be deprecated.
"""

from __future__ import absolute_import

try:
    from numpydoc import __version__ as numpydoc_version
except ImportError:  # __version__ only added in 0.7
    numpydoc_version = '0.0'

# this feature will be added to numpydoc in 0.7
if numpydoc_version < '0.7':

    from numpydoc import docscrape_sphinx

    # can't just replace the whole class, SphinxObjDoc complains, so need to
    # swap out just the method we want

    # record the original
    _str_examples_orig = docscrape_sphinx.SphinxDocString._str_examples

    # redefine the class method
    def _str_examples(self):
        examples_str = "\n".join(self['Examples'])
        if self.use_plots and 'plot::' not in examples_str and (
                'plot.show()' in examples_str or
                'plt.show()' in examples_str or
                'fig.show()' in examples_str):
            self['Examples'].insert(0, 'import matplotlib')
            lines = _str_examples_orig(self)
            lines.pop(lines.index('    import matplotlib'))
            return lines
        return _str_examples_orig(self)

    # replace the class method
    docscrape_sphinx.SphinxDocString._str_examples = _str_examples



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Multi-processed I/O utilities for `TimeSeries`
"""

from math import ceil

from six import string_types

from astropy.io.registry import read as io_read

from ...io.registry import get_read_format
from ...io.cache import (read_cache, cache_segments, file_segment, FILE_LIKE)
from ...segments import (Segment, SegmentList)
from ...utils.mp import multiprocess_with_queues
from ..core import TimeSeriesBaseList
from ..timeseries import TimeSeries
from ..statevector import (StateVector, StateVectorDict)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def read_from_cache(cache, channel, start=None, end=None, resample=None,
                    gap=None, pad=None, nproc=1, format=None, verbose=False,
                    **kwargs):
    """Read a `TimeSeries` from a cache of data files using multiprocessing

    The inner-workings are agnostic of data-type, but can only handle a
    single data type at a time.

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`, `str`
        cache of GWF frame files, or path to a LAL-format cache file
        on disk

    channel : :class:`~gwpy.detector.channel.Channel`, `str`
        data channel to read from frames

    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data

    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data

    resample : `float`, optional
        rate (samples per second) to resample

    format : `str`, optional
        name of data file format, e.g. ``gwf`` or ``hdf``.

    nproc : `int`, default: ``1``
        maximum number of independent frame reading processes, default
        is set to single-process file reading.

    gap : `str`, optional
        how to handle gaps in the cache, one of
        - 'ignore': do nothing, let the undelying reader method handle it
        - 'warn': do nothing except print a warning to the screen
        - 'raise': raise an exception upon finding a gap (default)
        - 'pad': insert a value to fill the gaps

    pad : `float`, optional
        value with which to fill gaps in the source data, only used if
        gap is not given, or `gap='pad'` is given

    Notes
    -----
    The number of independent processes spawned by this function can be
    calculated as ``min(maxprocesses, len(cache)//minprocesssize)``.

    Returns
    -------
    data : `~gwpy.timeseries.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    cls = kwargs.pop('target', TimeSeries)

    # open cache file
    if isinstance(cache, FILE_LIKE + string_types):
        cache = read_cache(cache)
    cache = type(cache)(cache)
    cache.sort(key=lambda e: e.segment)

    # get input format
    if kwargs.get('format', None) is None:
        kwargs['format'] = get_read_format(cls, cache[0].path, (), {})

    # handle resampling as an overlap for TimeSeries
    if not resample or cls in (StateVector, StateVectorDict):
        overlap = 0
    else:
        overlap = 4

    # get timing
    if start is None:  # start time of earliest file
        start = cache[0].segment[0]
    if end is None:  # end time of latest file
        end = cache[-1].segment[-1]
    span = Segment(start, end)
    cache = cache.sieve(segment=span.protract(overlap))
    cspan = Segment(cache[0].segment[0], cache[-1].segment[1])

    # define segments for multiprocessing
    segs = get_mp_cache_segments(cache, nproc, span, overlap=overlap)

    # define reader for each segment
    def _read_segment(segment):
        try:
            if segment[0] is None:
                pseg = segment
                subcache = cache
            else:
                pseg = segment.protract(overlap) & cspan  # processing segment
                subcache = cache.sieve(segment=pseg)
            data = io_read(cls, subcache, channel, start=pseg[0], end=pseg[1],
                           **kwargs)
            if resample:
                data = data.resample(resample)
            if overlap:  # crop out the overlap introduced for resampling
                return data.crop(*segment)
            return data
        except Exception as exc:  # pylint: disable=broad-except
            if nproc == 1:  # raise now
                raise
            return exc  # will be raised after multiprocessing closes

    # multi-process
    chunks = multiprocess_with_queues(nproc, _read_segment, segs,
                                      raise_exceptions=True, verbose=verbose)

    # flatten dicts
    if issubclass(cls, dict):
        out = cls()
        while chunks:
            new = chunks.pop(0)
            out.append(new, gap=gap, pad=pad)
            del new
        return out

    # flatten series
    list_ = TimeSeriesBaseList(*chunks)
    return list_.join(pad=pad, gap=gap)


def get_mp_cache_segments(cache, nproc, span, overlap=0):
    """Determine segments in which to multiprocess a cache
    """
    # no data
    if not cache:
        return [(None, None)]

    # minimum 1 file per process
    numf = len(cache)
    nproc = min(nproc, numf)

    # single process
    if nproc == 1:
        return SegmentList([span])

    # -- actual multiprocessing

    datasegs = cache_segments(cache) & SegmentList([span])
    secsperproc = abs(datasegs) / float(nproc)

    # copy cache and sort by start time
    cache = type(cache)(cache)
    cache.sort(key=lambda e: e.segment[0])

    segs = SegmentList()

    for seg in datasegs:
        # get files only for this data segment
        subcache = cache.sieve(segment=(seg.protract(overlap) & span))

        def pop_segment():
            return file_segment(subcache.pop(0)) & span

        # build process segment by looping through files in order
        while subcache:
            # create new seg by adding file durations until big enough
            newseg = pop_segment()
            while subcache and abs(newseg) < secsperproc:
                newseg += pop_segment()

            # if one file remains and it's shorter than a full segment,
            # add it to the current proces group; this should prevent
            # trying to read less than one sample from a GWF file, for example
            if (len(subcache) == 1 and
                    abs(cache_segments(subcache)) < secsperproc):
                newseg += pop_segment()

            segs.append(newseg)

    return segs



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for multi-processing
"""

import sys
from multiprocessing import (Queue, Process, Value)
from operator import itemgetter
from functools import partial

from six.moves import StringIO

from astropy.utils.console import (ProgressBar, ProgressBarOrSpinner,
                                   color_print)

from .misc import null_context


def process_in_out_queues(func, q_in, q_out):
    """Iterate through a Queue, call, ``func`, and Queue the result

    Parameters
    ----------
    func : `callable`
        any function that can take an element of the input `Queue` as
        the only argument

    q_in : `multiprocessing.queue.Queue`
        the input `Queue`

    q_out : `multiprocessing.queue.Queue`
        the output `Queue`

    Notes
    -----
    To close the input `Queue`, add ``(None, None)` as the last item
    """
    while True:
        # pick item out of input wqueue
        idx, arg = q_in.get()
        if idx is None:  # sentinel
            break
        # execute method and put the result in the output queue
        q_out.put((idx, func(arg)))


def multiprocess_with_queues(nproc, func, inputs, raise_exceptions=False,
                             verbose=False, **progress_kw):
    """Map a function over a list of inputs using multiprocess

    This essentially duplicates `multiprocess.map` but allows for
    arbitrary functions (that aren't necessarily importable)

    Parameters
    ----------
    nproc : `int`
        number of processes to use, if ``1`` is given, the current process
        is used, and no child processes are forked

    func : `callable`
        the function to call in each iteration, should take a single
        argument that is the next element from ``inputs``

    inputs : `iterable`
        iterable (e.g. `list`) of inputs, each element of which is
        passed to ``func`` in one of the child processes

    raise_exceptions : `bool`, optional, default: `False`
        if `True`, if the output of any ``func(input)`` calls is
        an `Exception`, it will be raised directly, otherwise the outputs
        are simply returned. This is useful if ``func`` can be built to
        detect exceptions and return then, rather than raising, so that
        child processes don't hang in multiprocessing when errors occur

    verbose : `bool`, `str`, optional
        if `True`, print progress to the console as a bar, pass a
        `str` to customise the heading for the progress bar, default: `False`,
        (default heading ``'Processing:'`` if ``verbose=True`)

    Returns
    -------
    outputs : `list`
        the `list` of results from calling ``func(x)`` for each element
        of ``inputs``
    """
    # handle verbose printing with a progress bar
    if verbose:
        total = len(inputs) if isinstance(inputs, (list, tuple)) else None
        if verbose is True:
            verbose = 'Processing:'
        bar = _MultiProgressBarOrSpinner(total, verbose, file=sys.stdout)

        def _inner_func(in_):  # update progress bar on each iteration
            res = func(in_)
            bar.update(None)
            return res

    else:
        # or don't
        bar = null_context()
        _inner_func = func

    # -------------------------------------------

    # shortcut single process
    if nproc == 1:
        with bar:
            return list(map(_inner_func, inputs))

    # create input and output queues
    q_in = Queue()
    q_out = Queue()

    # create child processes and start
    proclist = [Process(target=process_in_out_queues,
                        args=(_inner_func, q_in, q_out)) for _ in range(nproc)]

    with bar:

        for proc in proclist:
            proc.daemon = True
            proc.start()

        # populate queue (no need to block in serial put())
        sent = list(map(lambda x: q_in.put(x, block=False), enumerate(inputs)))
        for _ in range(nproc):  # add sentinel for each process
            q_in.put((None, None))

        # get results
        res = [q_out.get() for _ in range(len(sent))]

        # close processes and unwrap results
        for proc in proclist:
            proc.join()

    results = [out for i, out in sorted(res, key=itemgetter(0))]

    # raise exceptions
    if raise_exceptions:
        for exc in results:
            if isinstance(exc, Exception):
                raise exc

    return results


# -- progress bars with counter locking ---------------------------------------
# this is very much a hack, and should be pushed upstream to astropy proper

class _MultiProgressBar(ProgressBar):
    def update(self, value=None):
        # Update self.value
        if value is None:
            self._current_value.value += 1
        else:
            self._current_value = Value('i', value)
        value = self._current_value.value

        # Choose the appropriate environment
        if self._ipython_widget:
            self._update_ipython_widget(value)
        else:
            self._update_console(value)


class _MultiProgressBarOrSpinner(ProgressBarOrSpinner):
    def __init__(self, total, msg, color='default', file=None):
        if total is None:
            super(_MultiProgressBarOrSpinner, self).__init__(
                total, msg, color=color, file=file)
        else:
            self._is_spinner = False
            color_print(msg, color, file=file)
            self._obj = _MultiProgressBar(total, file=file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) Joseph Areeda (2017)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
#

""" Q-transform plots
"""

import re
from .cliproduct import CliProduct
import types
from pprint import pprint
from time import time


class Qtransform(CliProduct):
    """Plot the Q-transform (Omega)"""
    start_time = None
    qxfrm_args = None
    my_ts = None
    title_save = None

    def __init__(self):
        self.start_time = time()
        self.qxfrm_args = dict()
        super(Qtransform, self).__init__()

    def get_action(self):
        """Return the string used as "action" on command line."""
        return 'qtransform'

    def init_cli(self, parser):
        """Set up the argument list for this product"""

        self.arg_qxform(parser)
        self.arg_ax_linx(parser)
        self.arg_ax_ylf(parser)
        self.arg_ax_intlin(parser)
        self.arg_imag(parser)
        self.arg_plot(parser)
        return

    def post_arg(self, args):
        """Derive standard args from our weird ones
        :type args: Namespace with command line arguments
        """
        event = float(args.gps)
        search = int(args.search)

        start = int(event - search / 2)
        epoch = event
        args.start = [str(start)]
        args.epoch = ('%.3f' % epoch)
        args.duration = search
        args.chan = [[args.chan]]
        args.highpass = None
        args.lowpass = None
        self.verbose = args.verbose

    def get_ylabel(self, args):
        """Default text for y-axis label"""
        return 'Frequency (Hz)'

    def get_color_label(self):
        return 'Normalized energy'

    def get_max_datasets(self):
        """Q-transform only handles 1 at a time"""
        return 1

    def is_image(self):
        """This plot is image type"""
        return True

    def freq_is_y(self):
        """This plot puts frequency on the y-axis of the image"""
        return True

    def get_title(self):
        """Start of default super title, first channel is appended to it"""
        return 'Q-transform: '

    def gen_plot(self, args):
        """Generate the plot from time series and arguments"""
        self.is_freq_plot = False   # not fft based

        self.my_ts = self.timeseries[0]
        self.title2 = ''

        self.qxfrm_args['search'] = abs(self.my_ts.span) / 2.
        if args.qrange:
            self.qxfrm_args['qrange'] = (float(args.qrange[0]),
                                         float(args.qrange[1]))
            self.title2 += (' q-range [%.1f, %.1f], ' %
                            (self.qxfrm_args['qrange'][0],
                             self.qxfrm_args['qrange'][1]))
        if args.frange:
            self.qxfrm_args['frange'] = (float(args.frange[0]),
                                         float(args.frange[1]))

        if args.nowhiten:
            self.qxfrm_args['whiten'] = False
            self.title2 += 'not whitened, '
        else:
            self.title2 += 'whitened, '

        self.qxfrm_args['gps'] = float(args.gps)
        self.qxfrm_args['fres'] = 0.5
        self.qxfrm_args['tres'] = 0.002

        new_fs = float(args.sample_freq)
        cur_fs = self.my_ts.sample_rate.value

        if cur_fs > new_fs:
            self.my_ts = self.my_ts.resample(new_fs)
            self.title2 = (' %.0f resampled to %.0f Hz, ' %
                           (cur_fs, new_fs)) + self.title2
            self.log(3, 'Resampled input to %d Hz' % new_fs)
        else:
            new_fs = cur_fs
            self.title2 = (' %.0f Hz, ' % cur_fs) + self.title2

        prange = self.get_plot_range(args)
        epoch = float(args.epoch)
        self.qxfrm_args['outseg'] = (epoch-prange, epoch+prange)

        if self.verbose >= 3:
            print('Q-transform args:')
            pprint(self.qxfrm_args)

        self.result = self.my_ts.q_transform(**self.qxfrm_args)
        self.log(2, 'Result shape: %dx%d' %
                 (self.result.shape[0], self.result.shape[1]))

        self.pltargs = dict()
        if args.cmap:
            self.pltargs['cmap'] = args.cmap
        # weirdness because we allow 2 ways to specify intensity range
        imin = None
        imax = None
        if args.imin:
            imin = float(args.imin)
        if args.imax:
            imax = float(args.imax)
        if args.erange:
            imin = float(args.erange[0])
            imax = float(args.erange[1])

        if imin is not None:
            self.pltargs['vmin'] = imin
        if imax:
            self.pltargs['vmax'] = imax

        if self.verbose >= 3:
            print('Plot args:')
            pprint(self.pltargs)

        self.plot = self.result.plot(**self.pltargs)
        self.scaleText = 'Normalized energy'

        fmax = self.result.frequencies.max().value
        fmin = self.result.frequencies.min().value
        self.log(2, 'Frequency range of result: %.2f - %.2f' %
                 (fmin, fmax))
        self.title2 += (' calc f-range [%.1f, %.1f], ' %
                        (fmin, fmax))
        self.fmin = fmin
        self.fmax = fmax

        emin = self.result.min()
        emax = self.result.max()
        self.title2 += (' calc e-range [%.1f, %.1f] ' %
                        (emin, emax))

        self.title_save = self.title2
        self.title2 = (' Q = %.1f ' % self.result.q) + self.title_save
        self.plot_num = 0
        self.qx_plot_setup(args)

    def qx_plot_setup(self, args):
        """Next plot has different display (time) range"""
        prange = self.get_plot_range(args)
        epoch = float(args.epoch)
        args.xmin = ('%.3f' % (epoch - prange / 2))
        args.xmax = ('%.3f' % (epoch + prange / 2))

        args.out = ('%s/%s-%.3f-%05.2f.png' % (args.outdir,
                    re.sub(':', '-', self.timeseries[0].channel.name),
                    float(args.gps), prange))
        self.qxfrm_args['outseg'] = (epoch-prange, epoch+prange)

    def get_plot_range(self, args):
        prng = args.plot
        if isinstance(prng, types.ListType):
            prange = float(prng[self.plot_num])
        elif prng:
            prange = float(prng)
        else:
            prange = 0.5
            args.plot = '0.5'

        return prange

    def has_more_plots(self, args):
        """any ranges left to plot?"""
        self.plot_num += 1
        if not args.plot:
            ret = False
        else:
            ret = self.plot_num < len(args.plot)
        if not ret:
            run_time = time() - self.start_time
            self.log(2, 'Q-transform run time: %.1f  sec' % run_time)
        return ret

    def prep_next_plot(self, args):
        """Overridden because we may need multiple saves
        """
        self.qx_plot_setup(args)
        self.result = self.my_ts.q_transform(**self.qxfrm_args)
        self.plot = self.result.plot(**self.pltargs)
        self.title2 = (' Q = %.1f ' % self.result.q) + self.title_save



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Python version compatibilities
"""

import warnings
warnings.warn('The gwpy.utils.compat module has been deprecated and will '
              'be removed in an upcoming release, all objects provided '
              'by this module can be found in other libraries.',
              DeprecationWarning)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

from collections import OrderedDict  # pylint: disable=unused-import



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""tconvert: a utility to convert to and from GPS times.

This method is inspired by the original tconvert utility, written by
Peter Shawhan.
"""

import datetime
from decimal import Decimal

from dateutil import parser as dateparser

from astropy.units import Quantity

from . import (Time, LIGOTimeGPS)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['tconvert', 'to_gps', 'from_gps']


def tconvert(gpsordate='now'):
    """Convert GPS times to ISO-format date-times and vice-versa.

    Parameters
    ----------
    gpsordate : `float`, `astropy.time.Time`, `datetime.datetime`, ...
        input gps or date to convert, many input types are supported

    Returns
    -------
    date : `datetime.datetime` or `LIGOTimeGPS`
        converted gps or date

    Notes
    -----
    If the input object is a `float` or `LIGOTimeGPS`, it will get
    converted from GPS format into a `datetime.datetime`, otherwise
    the input will be converted into `LIGOTimeGPS`.

    Examples
    --------
    Integers and floats are automatically converted from GPS to
    `datetime.datetime`:

    >>> from gwpy.time import tconvert
    >>> tconvert(0)
    datetime.datetime(1980, 1, 6, 0, 0)
    >>> tconvert(1126259462.3910)
    datetime.datetime(2015, 9, 14, 9, 50, 45, 391000)

    while strings are automatically converted to `~gwpy.time.LIGOTimeGPS`:

    >>> to_gps('Sep 14 2015 09:50:45.391')
    LIGOTimeGPS(1126259462, 391000000)

    Additionally, a few special-case words as supported, which all return
    `~gwpy.time.LIGOTimeGPS`:

    >>> tconvert('now')
    >>> tconvert('today')
    >>> tconvert('tomorrow')
    >>> tconvert('yesterday')
    """
    # convert from GPS into datetime
    try:
        float(gpsordate)  # if we can 'float' it, then its probably a GPS time
    except (TypeError, ValueError):
        return to_gps(gpsordate)
    return from_gps(gpsordate)


def to_gps(t, *args, **kwargs):
    """Convert any input date/time into a `LIGOTimeGPS`.

    Any input object that can be cast as a `~astropy.time.Time`
    (with `str` going through the `datetime.datetime`) are acceptable.

    Parameters
    ----------
    t : `float`, `~datetime.datetime`, `~astropy.time.Time`, `str`
        the input time, any object that can be converted into a
        `LIGOTimeGPS`, `~astropy.time.Time`, or `~datetime.datetime`,
        is acceptable.

    *args, **kwargs
        other arguments to pass to pass to `~astropy.time.Time` if given

    Returns
    -------
    gps : `LIGOTimeGPS`
        the number of GPS seconds (non-integer) since the start of the
        epoch (January 6 1980).

    Raises
    ------
    TypeError
        if a `str` input cannot be parsed as a `datetime.datetime`.
    ValueError
        if the input cannot be cast as a `~astropy.time.Time` or
        `LIGOTimeGPS`.

    Examples
    --------
    >>> to_gps('Jan 1 2017')
    LIGOTimeGPS(1167264018, 0)
    >>> to_gps('Sep 14 2015 09:50:45.391')
    LIGOTimeGPS(1126259462, 391000000)

    >>> import datetime
    >>> to_gps(datetime.datetime(2017, 1, 1))
    LIGOTimeGPS(1167264018, 0)

    >>> from astropy.time import Time
    >>> to_gps(Time(57754, format='mjd'))
    LIGOTimeGPS(1167264018, 0)
    """
    # allow Time conversion to override type-checking
    if args or kwargs:
        return Time(t, *args, **kwargs).utc.gps
    # if lal.LIGOTimeGPS, just return it
    if isinstance(t, LIGOTimeGPS):
        return t
    # if Decimal, cast to LIGOTimeGPS and return
    if isinstance(t, Decimal):
        return LIGOTimeGPS(str(t))

    # or convert numeric string to float (e.g. '123.456')
    try:
        t = float(t)
    except (TypeError, ValueError):
        pass
    # or convert str into datetime.datetime
    if isinstance(t, str):
        t = str_to_datetime(t)
    # or convert tuple into datetime.datetime
    elif isinstance(t, (tuple, list)):
        t = datetime.datetime(*t)
    # and then into lal.LIGOTimeGPS or Time
    if isinstance(t, datetime.date):
        # try and use LAL, it's more reliable (possibly)
        try:
            from lal import UTCToGPS
        except ImportError:
            if not isinstance(t, datetime.datetime):
                t = datetime.datetime.combine(t, datetime.time.min)
            t = Time(t, scale='utc')
        else:
            gps = to_gps(UTCToGPS(t.timetuple()))
            if hasattr(t, 'microsecond'):
                return gps + t.microsecond * 1e-6
            return gps
    # and then into LIGOTimeGPS
    if isinstance(t, Time):
        return time_to_gps(t)
    # extract quantity to a float in seconds
    if isinstance(t, Quantity):
        t = t.to('second').value
    # if all else fails...
    return LIGOTimeGPS(t)


def from_gps(gps):
    """Convert a GPS time into a `datetime.datetime`.

    Parameters
    ----------
    gps : `LIGOTimeGPS`, `int`, `float`
        GPS time to convert

    Returns
    -------
    datetime : `datetime.datetime`
        ISO-format datetime equivalent of input GPS time

    Examples
    --------
    >>> from_gps(1167264018)
    datetime.datetime(2017, 1, 1, 0, 0)
    >>> from_gps(1126259462.3910)
    datetime.datetime(2015, 9, 14, 9, 50, 45, 391000)
    """
    try:
        gps = LIGOTimeGPS(gps)
    except (ValueError, TypeError):
        gps = LIGOTimeGPS(float(gps))
    try:
        from lal import GPSToUTC
    except ImportError:
        date = Time(gps.gpsSeconds, gps.gpsNanoSeconds * 1e-9,
                    format='gps', scale='utc').datetime
    else:
        date = datetime.datetime(*GPSToUTC(gps.gpsSeconds)[:6])
        date += datetime.timedelta(seconds=gps.gpsNanoSeconds * 1e-9)
    if float(gps).is_integer():
        return date.replace(microsecond=0)
    return date


def time_to_gps(t):
    """Convert a `Time` into `LIGOTimeGPS`.

    Parameters
    ----------
    t : `~astropy.time.Time`
        formatted `Time` object to convert

    Returns
    -------
    gps : `LIGOTimeGPS`
        Nano-second precision `LIGOTimeGPS` time
    """
    t = t.utc
    dt = t.datetime
    gps = t.gps
    # if datetime format has zero microseconds, force int(gps) to remove
    # floating point precision errors from gps
    if type(dt) is datetime.date or (  # pylint: disable=unidiomatic-typecheck
            isinstance(dt, datetime.datetime) and not dt.microsecond):
        return LIGOTimeGPS(int(gps))
    # use repr() to remove hidden floating point precision problems
    return LIGOTimeGPS(repr(gps))


def str_to_datetime(datestr):
    """Convert a `str` representing a datetime into a `datetime.datetime`.

    Parameters
    ----------
    datestr : `str`
        date-like string parseable by :meth:`dateutil.parser.parse`, or
        one of the following special cases

            - ``'now'`` : second precision for current time
            - ``'today'``
            - ``'tomorrow'``
            - ``'yesterday'``

    Returns
    -------
    date : `datetime.datetime`
        `datetime.datetime` version of the input ``datestr``

    Raises
    ------
    TypeError
        if ``datestr`` cannot be parsed by :meth:`dateutil.parser.parse`.
    """
    datestr = str(datestr).lower()
    if datestr == 'now':
        date = datetime.datetime.utcnow().replace(microsecond=0)
    elif datestr == 'today':
        date = datetime.date.today()
    elif datestr == 'tomorrow':
        today = datetime.date.today()
        date = today + datetime.timedelta(days=1)
    elif datestr == "yesterday":
        today = datetime.date.today()
        date = today - datetime.timedelta(days=1)
    else:
        try:
            date = dateparser.parse(datestr)
        except (ValueError, TypeError) as exc:
            exc.args = ("Cannot parse date string %r: %s"
                        % (datestr, exc.args[0]),)
            raise
    return date



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the GWpy package
"""

# ignore all invalid names (pylint isn't good at looking at executables)
# pylint: disable=invalid-name

from __future__ import print_function

import sys
from distutils.version import LooseVersion

from setuptools import (setup, find_packages,
                        __version__ as setuptools_version)

import versioneer
from setup_utils import (CMDCLASS, get_setup_requires, get_scripts)

__version__ = versioneer.get_version()

PEP_508 = LooseVersion(setuptools_version) >= '36.2'

# read description
with open('README.rst', 'rb') as f:
    longdesc = f.read().decode().strip()

# -- dependencies -------------------------------------------------------------

# build dependencies
setup_requires = get_setup_requires()

# runtime dependencies
dependencies = {
    'six': '>= 1.5',
    'python-dateutil': None,
    'numpy': '>= 1.7.1',
    'scipy': '>= 0.12.1',
    'matplotlib': '>= 1.2.0',
    'astropy': '>= 1.1.1',
    'lscsoft-glue': '>= 1.55.2',
}

# include fancy dependencies
if PEP_508:
    # exclude astropy >= 3.0 on python2.7
    dependencies['astropy '] = "{} ; python_version >= '3'".format(
        dependencies['astropy'])
    dependencies['astropy'] += " ; python_version < '3'"

    # exclude matplotlib 2.1.[01] (see matplotlib/matplotlib#10003)
    dependencies['matplotlib'] += ', !=2.1.0, !=2.1.1'

# unwrap dependencies into simple list
install_requires = ['{0} {1}'.format(pkg.strip(), ver or '') for
                    pkg, ver in dependencies.items()]

# test for LAL
try:
    import lal  # pylint: disable=unused-import
except ImportError as e:
    install_requires.append('ligotimegps >= 1.2.1')

# enum34 required for python < 3.4
try:
    import enum  # pylint: disable=unused-import
except ImportError:
    install_requires.append('enum34')

# define extras
extras_require = {
    'hdf5': ['h5py>=1.3'],
    'root': ['root_numpy'],
    'segments': ['dqsegdb'],
    'hacr': ['pymysql'],
    'docs': ['sphinx>=1.6.1', 'numpydoc', 'sphinx-bootstrap-theme>=0.6',
             'sphinxcontrib-programoutput', 'sphinx-automodapi',
             'requests'],
}

# define 'all' as the intersection of all extras
extras_require['all'] = set(p for extra in extras_require.values()
                            for p in extra)

# test dependencies
tests_require = [
    'pytest>=3.1',
    'freezegun',
    'sqlparse',
    'bs4',
]
if sys.version < '3':
    tests_require.append('mock')

# -- run setup ----------------------------------------------------------------

setup(
    # metadata
    name='gwpy',
    provides=['gwpy'],
    version=__version__,
    description="A python package for gravitational-wave astrophysics",
    long_description=longdesc,
    author='Duncan Macleod',
    author_email='duncan.macleod@ligo.org',
    license='GPLv3',
    url='https://gwpy.github.io/',

    # package content
    packages=find_packages(),
    scripts=get_scripts(),
    include_package_data=True,

    # dependencies
    cmdclass=CMDCLASS,
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,

    # classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
)
